from __future__ import annotations

import traceback
from logging import DEBUG as LVL_DEBUG
from logging import INFO as LVL_INFO
from typing import TYPE_CHECKING, Any, Final

import smllib
from smllib import SmlStreamReader
from smllib.errors import CrcError
from typing_extensions import Self

from sml2mqtt import CONFIG
from sml2mqtt.__log__ import get_logger
from sml2mqtt.const import EnhancedSmlFrame
from sml2mqtt.errors import ObisIdForConfigurationMappingNotFoundError, Sml2MqttExceptionWithLog
from sml2mqtt.mqtt import BASE_TOPIC
from sml2mqtt.sml_device.sml_devices import ALL_DEVICES
from sml2mqtt.sml_value import SmlValues

from .device_status import DeviceStatus
from .setup_device import setup_device
from .stream_reader_group import StreamReaderGroup, create_stream_reader_group
from .watchdog import Watchdog


if TYPE_CHECKING:
    from collections.abc import Callable

    from sml2mqtt.const import SourceProto

# -------------------------------------------------------------------------------------------------
# Dirty:
# Replace the class SmlFrame which is returned by SmlStreamReader with EnhancedSmlFrame
assert hasattr(smllib.reader, 'SmlFrame')
setattr(smllib.reader, 'SmlFrame', EnhancedSmlFrame)


class SmlDevice:
    def __init__(self, name: str) -> None:
        self._name: Final = name
        self._source: SourceProto | None = None
        self.status: DeviceStatus = DeviceStatus.STARTUP

        self.watchdog: Final = Watchdog(self)
        self.stream_reader: StreamReaderGroup | SmlStreamReader = create_stream_reader_group()

        self.log = get_logger(self.name)
        self.log_status = self.log.getChild('status')

        self.mqtt_device: Final = BASE_TOPIC.create_child(self.name)
        self.mqtt_status: Final = self.mqtt_device.create_child('status')

        self.device_id: str | None = None
        self.sml_values: Final = SmlValues()

        self.frame_handler: Callable[[EnhancedSmlFrame], Any] = self.process_first_frame

    @property
    def name(self) -> str:
        return self._name

    def set_source(self, source: SourceProto) -> Self:
        assert self._source is None, self._source
        self._source = source
        return self

    async def start(self) -> None:
        if self._source is not None:
            self._source.start()
        self.watchdog.start()

    async def cancel_and_wait(self) -> None:
        if self._source is not None:
            await self._source.cancel_and_wait()
        await self.watchdog.cancel_and_wait()

    def set_status(self, new_status: DeviceStatus) -> bool:
        if (old_status := self.status) == new_status or old_status.is_shutdown_status():
            return False

        self.status = new_status

        # Don't log toggling between CRC_ERROR and OK. Only log if new status is not OK
        level = LVL_INFO
        if new_status is DeviceStatus.CRC_ERROR:
            level = LVL_DEBUG
        elif old_status is DeviceStatus.CRC_ERROR:
            if new_status is DeviceStatus.OK:
                level = LVL_DEBUG
            else:
                # Log old status if new status is not OK
                self.log_status.log(level, f'Old status {old_status:s}')

        self.log_status.log(level, f'{new_status:s}')
        self.mqtt_status.publish(new_status.value)

        ALL_DEVICES.check_status()
        return True

    def on_source_data(self, data: bytes) -> None:
        frame = None    # type: EnhancedSmlFrame | None

        try:
            self.watchdog.feed()
            self.stream_reader.add(data)

            try:
                if (frame := self.stream_reader.get_frame()) is None:
                    return None
            except CrcError as e:
                self.log.debug(f'Crc error: {e.crc_calc} != {e.crc_msg}')
                self.set_status(DeviceStatus.CRC_ERROR)
                return None

            # Process Frame
            self.frame_handler(frame)
        except Exception as e:
            # dump frame if possible
            if frame is not None:
                for line in frame.get_frame_str():
                    self.log.info(line)

            self.on_error(e)

    def on_error(self, e: Exception, *, show_traceback: bool = True) -> None:
        self.log.debug(f'Exception {type(e)}: "{e}"')

        # Log exception
        if isinstance(e, Sml2MqttExceptionWithLog):
            e.log_msg(self.log)
        else:
            if show_traceback:
                for line in traceback.format_exc().splitlines():
                    self.log.error(line)
            else:
                self.log.error(f'{type(e)}: {e}')

        # Signal that an error occurred
        self.set_status(DeviceStatus.ERROR)
        return None

    def on_source_failed(self, reason: str) -> None:
        self.log.error(f'Source failed: {reason}')
        self.set_status(DeviceStatus.SOURCE_FAILED)

    def on_timeout(self) -> None:
        self.set_status(DeviceStatus.MSG_TIMEOUT)

    def process_frame(self, frame: EnhancedSmlFrame) -> None:

        frame_values = frame.get_frame_values(self.log)

        self.sml_values.process_frame(frame_values)

        # There was no Error -> OK
        self.set_status(DeviceStatus.OK)

    def setup_values_from_frame(self, frame: EnhancedSmlFrame) -> None:
        if not isinstance(self.stream_reader, SmlStreamReader):
            self.stream_reader = self.stream_reader.get_reader()
            _func_name = self.stream_reader.crc_func.__name__
            crc_func_name = getattr(self.stream_reader.crc_func, '__module__', _func_name).rsplit('.', 1)[-1]
            self.log.debug(f'Using crc {crc_func_name:s}')

        frame_values = frame.get_frame_values(self.log)

        # search frame and see if we get a match
        for search_obis in CONFIG.general.device_id_obis:
            if (obis_value := frame_values.get_value(search_obis)) is not None:
                self.log.debug(f'Found obis id {search_obis:s} in the sml frame')
                self.device_id = str(obis_value.get_value())
                break
        else:
            searched = ', '.join(CONFIG.general.device_id_obis)
            self.log.error(f'Found none of the following obis ids in the sml frame: {searched:s}')
            raise ObisIdForConfigurationMappingNotFoundError()

        # Search configuration to see if we have a special config for the device
        device_cfg = CONFIG.devices.get(self.device_id)
        if device_cfg is None:
            self.log.warning(f'No device found for {self.device_id:s}')
        else:
            self.log.debug(f'Device found for {self.device_id:s}')
        setup_device(self, frame_values, device_cfg, CONFIG.general)

        self.frame_handler = self.process_frame

    def process_first_frame(self, frame: EnhancedSmlFrame) -> None:

        try:
            self.setup_values_from_frame(frame)
        except Exception as e:
            self.set_status(DeviceStatus.SHUTDOWN)
            self.on_error(e)
            return None

        self.frame_handler(frame)
        return None

    def analyze_frame(self, frame: EnhancedSmlFrame) -> None:

        # log Frame and frame description
        for line in frame.get_analyze_str():
            self.log.info(line)

        # Setup and process the frame
        self.setup_values_from_frame(frame)

        # Log setup values
        for line in self.sml_values.describe():
            self.log.info(line)

        # process the frame
        self.frame_handler(frame)

        # shutdown
        self.log.info('')
        self.set_status(DeviceStatus.SHUTDOWN)
        return None



import ssl
from logging import Logger
from pathlib import Path
from typing import Any, Literal

from aiomqtt import TLSParameters
from easyconfig import BaseModel
from pydantic import Field, field_validator


def _get_ssl_consts(prefix: str) -> dict[str, int]:
    return {name.removeprefix(prefix): getattr(ssl, name) for name in dir(ssl) if name.startswith(prefix)}


def get_ssl_verify_values() -> dict[str, int]:
    return _get_ssl_consts('CERT_')


def get_ssl_version_values() -> dict[str, int]:
    return _get_ssl_consts('PROTOCOL_')


class MqttTlsOptions(BaseModel):
    insecure: bool | None = Field(
        None, description='Enable/disable server hostname verification when using SSL/TLS.'
    )

    ca_certs: str | None = Field(
        None, alias='ca certificates',
        description='Path to Certificate Authority (CA) certificate file in PEM or DER format'
    )

    cert_file: str | None = Field(
        None, alias='cert file', description='Path to PEM encoded client certificate file'
    )
    key_file: str | None = Field(
        None, alias='key file', description='Path to PEM encoded private keys file'
    )
    file_password: str | None = Field(
        None, alias='file password', description='Password to encrypt the cert file or the key file if needed'
    )

    cert_reqs: Literal[tuple(get_ssl_verify_values())] | None = Field(
        None, alias='certificate requirement',
        description='Certificate requirement that the client imposes on the broker.'
    )
    tls_version: Literal[tuple(get_ssl_version_values())] | None = Field(
        None, alias='tls version', description='The version of the SSL/TLS protocol to be used.'
    )
    ciphers: str | None = Field(
        None, description='Which encryption ciphers are allowable for the connection'
    )

    @field_validator('ca_certs', 'cert_file', 'key_file')
    @classmethod
    def _ensure_file_exists(cls, v: str | None):
        if v is None:
            return None

        if not isinstance(v, str):
            raise TypeError()

        if not Path(v).is_file():
            raise FileNotFoundError(v)

        return v

    def get_client_kwargs(self, log: Logger) -> dict[str, Any]:
        cert_reqs = get_ssl_verify_values()[self.cert_reqs] if self.cert_reqs is not None else None
        tls_version = get_ssl_version_values()[self.tls_version] if self.tls_version is not None else None

        if self.insecure:
            log.warning('Verification of server hostname in server certificate disabled! '
                        'Use this only for testing, not for a real system!')

        return {
            'tls_insecure': self.insecure,
            'tls_params': TLSParameters(
                ca_certs=self.ca_certs,
                certfile=self.cert_file,
                keyfile=self.key_file,
                cert_reqs=cert_reqs,
                tls_version=tls_version,
                ciphers=self.ciphers,
                keyfile_password=self.file_password
            )
        }



from collections.abc import Generator
from typing import Final

from typing_extensions import override

from sml2mqtt.errors import RequiredObisValueNotInFrameError
from sml2mqtt.sml_value.base import SmlValueInfo, ValueOperationBase


class NegativeOnEnergyMeterWorkaroundOperation(ValueOperationBase):
    def __init__(self, meter_obis: str | None = None) -> None:
        self.meter_obis: Final[str] = '0100010800ff' if meter_obis is None else meter_obis

    @override
    def process_value(self, value: float | None, info: SmlValueInfo) -> float | None:
        if value is None:
            return None

        if (meter := info.frame.get_value(self.meter_obis)) is None:
            raise RequiredObisValueNotInFrameError(self.meter_obis)

        status = meter.status
        if not isinstance(status, int):
            msg = f'Energy Meter status {self.meter_obis:s} is not a valid int: {status} ({type(status)})'
            raise TypeError(msg)

        negative = status & 0x20
        if negative:
            value *= -1

        return value

    def __repr__(self) -> str:
        return f'<NegativeOnEnergyMeterWorkaround: {self.meter_obis:s} at 0x{id(self):x}>'

    @override
    def describe(self, indent: str = '') -> Generator[str, None, None]:
        yield f'{indent:s}- Negative On Status Of Energy Meter {self.meter_obis:s}'



from __future__ import annotations

from asyncio import Event, TimeoutError, wait_for
from typing import TYPE_CHECKING, Final

from ..const import DeviceTask


if TYPE_CHECKING:
    from .sml_device import SmlDevice


class Watchdog:
    def __init__(self, device: SmlDevice) -> None:
        self._timeout: float = -1
        self.device: Final = device

        self._event: Final = Event()
        self._task: Final  = DeviceTask(device, self._wd_task, name=f'Watchdog Task {self.device.name:s}')

    def start(self) -> None:
        self._task.start()

    def cancel(self) -> None:
        self._task.cancel()

    async def cancel_and_wait(self):
        return await self._task.cancel_and_wait()

    def set_timeout(self, timeout: float):
        if timeout < 0.1:
            raise ValueError()
        self._timeout = timeout
        return self

    def feed(self) -> None:
        self._event.set()

    async def _wd_task(self) -> None:
        make_call = True
        while True:
            self._event.clear()

            try:
                await wait_for(self._event.wait(), self._timeout)
                make_call = True
                continue
            except TimeoutError:
                pass

            # callback only once!
            if not make_call:
                continue
            make_call = False

            try:
                self.device.on_timeout()
            except Exception as e:
                self.device.on_error(e)



from __future__ import annotations

from typing import TYPE_CHECKING

from sml2mqtt.mqtt import BASE_TOPIC, check_for_duplicate_topics
from sml2mqtt.sml_value import SmlValue
from sml2mqtt.sml_value.operations import (
    FactorOperation,
    OffsetOperation,
    OnChangeFilterOperation,
    OrOperation,
    RangeFilterOperation,
    RefreshActionOperation,
    RoundOperation,
    SequenceOperation,
    SkipZeroMeterOperation,
)
from sml2mqtt.sml_value.setup_operations import setup_operations


if TYPE_CHECKING:
    import logging

    from sml2mqtt.config.config import GeneralSettings
    from sml2mqtt.config.device import SmlDeviceConfig
    from sml2mqtt.const import SmlFrameValues
    from sml2mqtt.sml_device import SmlDevice
    from sml2mqtt.sml_value.base import OperationContainerBase, ValueOperationBase


def has_operation_type(obj: OperationContainerBase, *ops: type[ValueOperationBase],
                       is_of: bool = True) -> ValueOperationBase | None:
    for op in obj.operations:
        if isinstance(op, ops) == is_of:
            return op
        if (isinstance(op, (OrOperation, SequenceOperation)) and
                (ret := has_operation_type(op, is_of=is_of)) is not None):
            return ret
    return None


def _create_default_transformations(log: logging.Logger, sml_value: SmlValue, frame: SmlFrameValues,
                                    general_cfg: GeneralSettings) -> None:

    op_count = len(sml_value.operations)

    if (entry := frame.get_value(sml_value.obis)) is not None and entry.unit == 30:
        if general_cfg.wh_in_kwh:
            if op := has_operation_type(sml_value, FactorOperation):
                log.debug(f'Found {op.__class__.__name__:s} - skip creating default factor')
            else:
                sml_value.insert_operation(FactorOperation(1 / 1000))

        # If the user has created something for the meter we don't skip it
        if not op_count:
            sml_value.insert_operation(SkipZeroMeterOperation())


def _create_default_filters(log: logging.Logger, sml_value: SmlValue, general_cfg: GeneralSettings):
    if op := has_operation_type(
        sml_value,
        FactorOperation, OffsetOperation, RoundOperation, RangeFilterOperation, SkipZeroMeterOperation,
        is_of=False
    ):
        log.debug(f'Found {op.__class__.__name__:s} - skip creating default filters')
        return None

    log.info(f'No filters found for {sml_value.obis}, creating default filters')

    sml_value.add_operation(OnChangeFilterOperation())
    sml_value.add_operation(RefreshActionOperation(general_cfg.republish_after))


def setup_device(device: SmlDevice, frame: SmlFrameValues, cfg: SmlDeviceConfig | None, general_cfg: GeneralSettings) -> None:
    mqtt_device = device.mqtt_device
    skip_default_setup = set()

    mqtt_device.set_topic(device.device_id)

    if cfg is not None:
        # mqtt of device
        mqtt_device.set_config(cfg.mqtt)
        device.mqtt_status.set_config(cfg.status)

        skipped_obis = set(cfg.skip)
        if not general_cfg.report_device_id and device.device_id is not None:
            skipped_obis.update(general_cfg.device_id_obis)

        device.sml_values.set_skipped(*skipped_obis)
        skip_default_setup.update(skipped_obis)

        for value_cfg in cfg.values:
            obis = value_cfg.obis
            skip_default_setup.add(obis)

            if obis in cfg.skip:
                device.log.warning(f'Config for {obis:s} found but {obis:s} is also marked to be skipped')
            if obis not in frame.obis_ids():
                device.log.warning(f'Config for {obis:s} found but {obis:s} was not reported by the frame')

            sml_value = SmlValue(
                obis,
                mqtt_device.create_child(topic_fragment=obis).set_config(value_cfg.mqtt)
            )
            device.sml_values.add_value(sml_value)

            setup_operations(sml_value, value_cfg)
            _create_default_transformations(device.log, sml_value, frame, general_cfg)
            _create_default_filters(device.log, sml_value, general_cfg)
    else:
        # No config found -> ignore defaults
        device.sml_values.set_skipped(*general_cfg.device_id_obis)

    # Create default for not
    for obis, _ in frame.items(skip=skip_default_setup):    # noqa: PERF102
        sml_value = SmlValue(obis, mqtt_device.create_child(topic_fragment=obis))
        device.sml_values.add_value(sml_value)

        _create_default_transformations(device.log, sml_value, frame, general_cfg)
        _create_default_filters(device.log, sml_value, general_cfg)

    # Check for duplicate MQTT topics
    check_for_duplicate_topics(BASE_TOPIC)



from __future__ import annotations

import logging
import traceback
from asyncio import CancelledError, Event, current_task
from asyncio import Task as asyncio_Task
from asyncio import create_task as asyncio_create_task
from typing import TYPE_CHECKING, Final


if TYPE_CHECKING:
    from collections.abc import Awaitable, Callable, Coroutine

    from sml2mqtt.const.protocols import DeviceProto


TASKS: Final[set[asyncio_Task]] = set()

log = logging.getLogger('sml.tasks')


def create_task(coro: Coroutine, *, name: str | None = None):
    task = asyncio_create_task(coro, name=name)

    TASKS.add(task)
    task.add_done_callback(TASKS.discard)
    return task


async def wait_for_tasks() -> None:

    while True:
        for task in TASKS.copy():
            if task.done():
                continue

            # these are the asyncio tasks. Exceptions are handled either in Task or DeviceTask,
            # so we can not await the tasks here because that would raise the Exception.
            # That's why we use an event to signal that the task is done
            event = Event()
            task.add_done_callback(lambda x: event.set())
            await event.wait()
            break

        else:
            break

    log.debug('All tasks done')


class Task:
    def __init__(self, coro: Callable[[], Awaitable], *, name: str) -> None:
        self._coro: Final = coro
        self._name: Final = name

        self._task: asyncio_Task | None = None

    @property
    def is_running(self) -> bool:
        if (task := self._task) is None or task.cancelled():  # noqa: SIM103
            return False
        return True

    def start(self) -> None:
        if not self.is_running:
            self._task = create_task(self._wrapper(), name=self._name)

    def cancel(self) -> asyncio_Task | None:
        if (task := self._task) is None:
            return None

        task.cancel()
        return task

    async def cancel_and_wait(self) -> bool:
        if (task := self.cancel()) is None:
            return False

        try:  # noqa: SIM105
            await task
        except CancelledError:
            pass
        return True

    async def _wrapper(self) -> None:
        task = current_task()

        try:
            await self._coro()
        except Exception as e:
            self.process_exception(e)
        except CancelledError:
            pass
        finally:
            if task is self._task:
                self._task = None

            log.debug(f'{self._name:s} finished!')

    def process_exception(self, e: Exception) -> None:
        log.error(f'Error in {self._name:s}')
        for line in traceback.format_exc().splitlines():
            log.error(line)


class DeviceTask(Task):
    def __init__(self, device: DeviceProto, coro: Callable[[], Coroutine], *, name: str) -> None:
        super().__init__(coro, name=name)
        self._device: Final = device

    def process_exception(self, e: Exception) -> None:
        super().process_exception(e)
        self._device.on_source_failed(f'Task crashed: {e.__class__.__name__}')



from collections import deque
from collections.abc import Sequence
from datetime import timedelta
from typing import Final, TypeAlias


DurationType: TypeAlias = timedelta | float | int


def get_duration(obj: DurationType) -> int | float:
    if isinstance(obj, timedelta):
        return int(obj.total_seconds()) if not obj.microseconds else obj.total_seconds()
    if not isinstance(obj, (int, float)):
        raise TypeError()
    return obj


class TimeSeries:
    __slots__ = ('period', 'times', 'values', 'is_full', 'wait_for_data')

    def __init__(self, period: DurationType, wait_for_data: bool = False) -> None:
        self.wait_for_data: Final = wait_for_data
        self.period: Final = get_duration(period)
        self.times: Final[deque[float]] = deque()
        self.values: Final[deque[float]] = deque()

        self.is_full: bool = False

    def clear(self) -> None:
        self.is_full = False
        self.times.clear()
        self.values.clear()

    def add_value(self, value: int | float | None, timestamp: float) -> None:
        start = timestamp - self.period

        if value is not None:
            self.times.append(timestamp)
            self.values.append(value)

        if not self.is_full and self.times and self.times[0] <= start:
            self.is_full = True

        try:
            while self.times[1] <= start:
                self.times.popleft()
                self.values.popleft()
        except IndexError:
            pass

    def get_values(self) -> Sequence[float] | None:
        if not self.values or self.wait_for_data and not self.is_full:
            return None
        return tuple(self.values)

    def get_value_duration(self, timestamp: float) -> Sequence[tuple[float, float]] | None:
        if not self.values or self.wait_for_data and not self.is_full:
            return None

        start_of_interval = timestamp - self.period

        start = self.times[0]
        if start <= start_of_interval:
            start = start_of_interval

        stop = timestamp
        value = self.values[0]

        ret = []
        for i in range(1, len(self.values)):
            stop = self.times[i]
            ret.append((value, stop - start))
            start = stop
            value = self.values[i]

        ret.append((value, stop - start))
        return ret



from enum import Enum


class DeviceStatus(str, Enum):
    STARTUP = 'STARTUP'
    SHUTDOWN = 'SHUTDOWN'

    SOURCE_FAILED = 'SOURCE_FAILED'
    MSG_TIMEOUT = 'MESSAGE_TIMEOUT'
    CRC_ERROR = 'CRC_ERROR'
    ERROR = 'ERROR'

    OK = 'OK'

    def is_shutdown_status(self) -> bool:
        return self.value in (DeviceStatus.ERROR, DeviceStatus.SOURCE_FAILED, DeviceStatus.SHUTDOWN)

    def __str__(self) -> str:
        return self.value



from __future__ import annotations

import asyncio
import logging
from asyncio import Protocol
from time import monotonic
from typing import TYPE_CHECKING, Final

from serial_asyncio import SerialTransport, create_serial_connection

from sml2mqtt.__log__ import get_logger
from sml2mqtt.const import DeviceTask


if TYPE_CHECKING:
    from sml2mqtt.config.inputs import SerialSourceSettings
    from sml2mqtt.const import DeviceProto


log = get_logger('serial')


class SerialSource(Protocol):
    @classmethod
    async def create(cls, device: DeviceProto, settings: SerialSourceSettings) -> SerialSource:
        transport, protocol = await create_serial_connection(
            asyncio.get_event_loop(),
            lambda: cls(device, settings.url),
            url=settings.url,
            baudrate=settings.baudrate, parity=settings.parity,
            stopbits=settings.stopbits, bytesize=settings.bytesize
        )  # type: SerialTransport, SerialSource

        return protocol

    def __init__(self, device: DeviceProto, url: str) -> None:
        super().__init__()

        self.url: Final = url
        self.device: Final = device

        self.transport: SerialTransport | None = None

        self._task: Final = DeviceTask(device, self._chunk_task, name=f'Serial Task {self.device.name:s}')

        self.last_read: float | None = 0.0

    def start(self) -> None:
        self._task.start()

    async def cancel_and_wait(self):
        return await self._task.cancel_and_wait()

    def connection_made(self, transport: SerialTransport) -> None:
        self.transport = transport
        log.debug(f'Port {self.url:s} successfully opened')

        # so we can read bigger chunks at once in case someone uses a higher baudrate
        self.transport._max_read_size = 10_240

    def connection_lost(self, exc: Exception | None) -> None:

        lvl = logging.INFO
        ex_str = ''
        if exc is not None:
            ex_str = f': {exc}'
            lvl = logging.ERROR

        log.log(lvl, f'Port {self.url:s} was closed{ex_str:s}')
        self.device.on_source_failed(f'Connection to port {self.url:s} lost')

    def data_received(self, data: bytes) -> None:
        self.transport.pause_reading()

        self.last_read = monotonic()
        self.device.on_source_data(data)

    async def _chunk_task(self) -> None:
        interval = 0.2

        while True:
            await asyncio.sleep(interval)

            if self.last_read is not None:
                diff_to_interval = interval - (monotonic() - self.last_read)
                self.last_read = None
                if diff_to_interval >= 0.001:
                    await asyncio.sleep(diff_to_interval)

            # safe to be called multiple times in a row
            self.transport.resume_reading()




from sml2mqtt.config.inputs import HttpSourceSettings, SerialSourceSettings
from sml2mqtt.const import DeviceProto, SourceProto


async def create_source(device: DeviceProto, settings: SerialSourceSettings | HttpSourceSettings) -> SourceProto:

    if isinstance(settings, SerialSourceSettings):
        from .serial import SerialSource
        return await SerialSource.create(device, settings)
    if isinstance(settings, HttpSourceSettings):
        from .http import HttpSource
        return await HttpSource.create(device, settings)

    msg = 'Unknown source input type'
    raise TypeError(msg)



from collections.abc import Generator
from time import monotonic
from typing import Final

from typing_extensions import override

from sml2mqtt.const import DurationType, get_duration
from sml2mqtt.sml_value.base import SmlValueInfo, ValueOperationBase
from sml2mqtt.sml_value.operations._helper import format_period


class RefreshActionOperation(ValueOperationBase):
    def __init__(self, every: DurationType) -> None:
        self.every: Final = get_duration(every)
        self.last_time: float = -1
        self.last_value: float | None = None

    @override
    def process_value(self, value: float | None, info: SmlValueInfo) -> float | None:
        if value is not None:
            self.last_value = value
            self.last_time = monotonic()
            return value

        if monotonic() - self.last_time < self.every:
            return None

        self.last_time = monotonic()
        return self.last_value

    def __repr__(self) -> str:
        return f'<RefreshAction: {self.every}s at 0x{id(self):x}>'

    @override
    def describe(self, indent: str = '') -> Generator[str, None, None]:
        yield f'{indent:s}- Refresh Action: {format_period(self.every)}'


class HeartbeatActionOperation(ValueOperationBase):
    def __init__(self, every: DurationType) -> None:
        self.every: Final = get_duration(every)
        self.last_time: float = -1_000_000_000
        self.last_value: float | None = None

    @override
    def process_value(self, value: float | None, info: SmlValueInfo) -> float | None:
        if value is not None:
            self.last_value = value

        if monotonic() - self.last_time < self.every:
            return None

        self.last_time = monotonic()
        return self.last_value

    def __repr__(self) -> str:
        return f'<HeartbeatAction: {self.every}s at 0x{id(self):x}>'

    @override
    def describe(self, indent: str = '') -> Generator[str, None, None]:
        yield f'{indent:s}- Heartbeat Action: {format_period(self.every)}'



from .connect_delay import DynDelay
from .mqtt import publish, start, wait_for_connect


# isort: split

from .mqtt_obj import BASE_TOPIC, MqttObj, check_for_duplicate_topics, patch_analyze, setup_base_topic



from easyconfig import AppBaseModel, BaseModel, create_app_config
from pydantic import Field

from .device import SmlDeviceConfig, SmlValueConfig
from .inputs import HttpSourceSettings, SerialSourceSettings
from .logging import LoggingSettings
from .mqtt import MqttConfig, OptionalMqttPublishConfig
from .types import LowerStr, ObisHex


class GeneralSettings(BaseModel):
    wh_in_kwh: bool = Field(
        True, description='Automatically convert Wh to kWh', alias='Wh in kWh'
    )
    republish_after: int = Field(
        120, description='Republish automatically after this time (if no other filter configured)',
        alias='republish after',
    )
    report_blank_energy_meters: bool = Field(
        False, description='Report blank energy meters (where the value is 0kwh)',
        alias='report blank energy meters', in_file=False
    )
    report_device_id: bool = Field(
        False, description='Report the device id even though it does never change',
        alias='report device id', in_file=False
    )
    device_id_obis: list[ObisHex] = Field(
        # 0100000009ff (1-0:0.0.9*255) : Geräteeinzelidentifikation
        # 0100600100ff (1-0:96.1.0*255): Produktionsnummer
        ['0100000009ff', '0100600100ff'],
        description='Additional OBIS fields for the serial number to configuration matching',
        alias='device id obis', in_file=False
    )
    crc: list[str] = Field(
        default=['x25', 'kermit'],
        description='Which crc algorithms are used to calculate the checksum of the smart meter',
        alias='crc', in_file=False
    )


class Settings(AppBaseModel):
    logging: LoggingSettings = Field(default_factory=LoggingSettings)
    mqtt: MqttConfig = Field(default_factory=MqttConfig)
    general: GeneralSettings = Field(default_factory=GeneralSettings)
    inputs: list[HttpSourceSettings | SerialSourceSettings] = Field(default_factory=list, discriminator='type')
    devices: dict[LowerStr, SmlDeviceConfig] = Field({}, description='Device configuration by ID or url',)


def default_config() -> Settings:
    # File defaults
    return Settings(
        inputs=[SerialSourceSettings(type='serial', url='COM1', timeout=6),
                SerialSourceSettings(type='serial', url='/dev/ttyS0', timeout=6), ],
        devices={
            'device_id_hex': SmlDeviceConfig(
                mqtt=OptionalMqttPublishConfig(topic='DEVICE_BASE_TOPIC'),
                status=OptionalMqttPublishConfig(topic='status'),
                skip={'00112233445566'},
                values=[
                    SmlValueConfig(
                        obis='00112233445566',
                        mqtt=OptionalMqttPublishConfig(topic='OBIS'),
                        operations=[
                            {'negative on energy meter status': True},
                            {'factor': 3}, {'offset': 100}, {'round': 2},
                            {'type': 'change filter'},
                            {'refresh action': 600}
                        ]
                    )
                ]
            )
        }
    )


CONFIG: Settings = create_app_config(Settings(), default_config)



import pytest

from sml2mqtt.mqtt import MqttObj, check_for_duplicate_topics


def test_topmost(monkeypatch) -> None:
    parent = MqttObj('base', 2, True).update()

    assert parent.topic == 'base'
    assert parent.qos == 2
    assert parent.retain is True

    parent.set_topic('base_new')
    assert parent.topic == 'base_new'
    assert parent.qos == 2
    assert parent.retain is True


def test_prefix_empty(monkeypatch) -> None:
    parent = MqttObj('', 2, True).update()
    child = parent.create_child('child')

    assert (child.topic, child.qos, child.retain) == ('child', 2, True)


def test_child_change(monkeypatch) -> None:
    parent = MqttObj('base', 2, True).update()
    child = parent.create_child('child')

    assert (child.topic, child.qos, child.retain) == ('base/child', 2, True)

    # QOS overwrite
    child.cfg.qos = 0
    child.update()
    assert (child.topic, child.qos, child.retain) == ('base/child', 0, True)

    # Retain overwrite
    child.cfg.retain = False
    child.update()
    assert (child.topic, child.qos, child.retain) == ('base/child', 0, False)

    # Load defaults from parent again
    child.cfg.qos = None
    child.cfg.retain = None
    child.update()
    assert (child.topic, child.qos, child.retain) == ('base/child', 2, True)

    # Parent changes - should reflect now on child
    parent.cfg.qos = 0
    parent.cfg.retain = False
    parent.update()
    assert (child.topic, child.qos, child.retain) == ('base/child', 0, False)


@pytest.mark.ignore_log_warnings
def test_check_for_duplicate_messages(caplog) -> None:
    parent = MqttObj('base', 2, True).update()
    parent.create_child('child')
    parent.create_child('child')

    check_for_duplicate_topics(parent)

    msg = '\n'.join(x.msg for x in caplog.records)

    assert msg == 'Topic "base/child" is already configured!'



from collections.abc import Generator
from typing import Final

from typing_extensions import override

from sml2mqtt.sml_value.base import SmlValueInfo, ValueOperationBase


class FactorOperation(ValueOperationBase):
    def __init__(self, factor: int | float) -> None:
        self.factor: Final = factor

    @override
    def process_value(self, value: float | None, info: SmlValueInfo) -> float | None:
        if value is None:
            return None
        return value * self.factor

    def __repr__(self) -> str:
        return f'<Factor: {self.factor} at 0x{id(self):x}>'

    @override
    def describe(self, indent: str = '') -> Generator[str, None, None]:
        yield f'{indent:s}- Factor: {self.factor}'


class OffsetOperation(ValueOperationBase):
    def __init__(self, offset: int | float) -> None:
        self.offset: Final = offset

    @override
    def process_value(self, value: float | None, info: SmlValueInfo) -> float | None:
        if value is None:
            return None
        return value + self.offset

    def __repr__(self) -> str:
        return f'<Offset: {self.offset} at 0x{id(self):x}>'

    @override
    def describe(self, indent: str = '') -> Generator[str, None, None]:
        yield f'{indent:s}- Offset: {self.offset}'


class RoundOperation(ValueOperationBase):
    def __init__(self, digits: int) -> None:
        self.digits: Final = digits if digits else None

    @override
    def process_value(self, value: float | None, info: SmlValueInfo) -> float | None:
        if value is None:
            return None

        if isinstance(value, int):
            return value
        return round(value, self.digits)

    def __repr__(self) -> str:
        return f'<Round: {self.digits if self.digits is not None else 0} at 0x{id(self):x}>'

    @override
    def describe(self, indent: str = '') -> Generator[str, None, None]:
        yield f'{indent:s}- Round: {self.digits if self.digits is not None else "integer"}'



from __future__ import annotations

from binascii import b2a_hex
from time import monotonic
from typing import TYPE_CHECKING, Any, Final

from smllib import SmlFrame


if TYPE_CHECKING:
    from collections.abc import Generator, Iterable
    from logging import Logger

    from smllib.sml import SmlListEntry


class EnhancedSmlFrame(SmlFrame):

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.timestamp: Final = monotonic()

    def get_frame_str(self):
        yield 'Received Frame'
        yield f' -> {b2a_hex(self.buffer)}'

    def get_analyze_str(self) -> Generator[str, None, None]:
        yield ''
        yield from self.get_frame_str()
        yield ''
        for obj in self.parse_frame():
            yield from obj.format_msg().splitlines()
        yield ''

    def get_frame_values(self, log: Logger) -> SmlFrameValues:

        # try shortcut, if that fails try parsing the whole frame
        try:
            sml_objs = self.get_obis()  # type: list[SmlListEntry]
        except Exception:
            log.info('get_obis failed - try parsing frame')

            sml_objs = []   # type: list[SmlListEntry]
            for msg in self.parse_frame():
                for val in getattr(msg.message_body, 'val_list', []):
                    sml_objs.append(val)

        return SmlFrameValues.create(self.timestamp, sml_objs)


class SmlFrameValues:
    @classmethod
    def create(cls, timestamp: float, values: Iterable[SmlListEntry]):
        c = cls(timestamp)
        for value in values:
            c.values[value.obis] = value
        return c

    def __init__(self, timestamp: float) -> None:
        self.timestamp: Final = timestamp
        self.values: Final[dict[str, SmlListEntry]] = {}

    def __getattr__(self, item: str) -> SmlListEntry:
        return self.values[item]

    def __len__(self) -> int:
        return len(self.values)

    def get_value(self, obis: str) -> SmlListEntry | None:
        return self.values.get(obis)

    def obis_ids(self) -> frozenset[str]:
        return frozenset(self.values)

    def items(self, skip: set[str]) -> Generator[tuple[str, SmlListEntry], Any, None]:
        for obis_id, value in self.values.items():
            if obis_id in skip:
                continue
            yield obis_id, value



from __future__ import annotations

from typing import TYPE_CHECKING, Final

from typing_extensions import override

from sml2mqtt.const import DateTimeFinder, get_now
from sml2mqtt.sml_value.base import SmlValueInfo, ValueOperationWithStartupBase


if TYPE_CHECKING:
    from collections.abc import Generator
    from datetime import datetime


class SupportsDateTimeAction(ValueOperationWithStartupBase):

    def __init__(self, dt_finder: DateTimeFinder, start_now: bool = True) -> None:
        self._dt_finder: Final = dt_finder
        self._next_reset: datetime = dt_finder.get_first_reset(start_now)

        if start_now or not self._dt_finder.enabled:
            self.enable_on_first_value()

    def after_next_reset(self, update: bool = True) -> bool:
        if not self._dt_finder.enabled:
            return False

        if (now := get_now()) >= self._next_reset:
            if update:
                self._next_reset = self._dt_finder.calc_next(now)
            return True

        return False

    @override
    def describe(self, indent: str = '') -> Generator[str, None, None]:
        if not self._dt_finder.enabled:
            yield f'{indent:s}    No resets'
            return None

        yield f'{indent:s}    Next resets:'
        yield f'{indent:s}     - {self._next_reset if self._next_reset >= get_now() else "now"}'

        next_dt = self._next_reset

        # for every condition we want to show two values
        samples = max(self._dt_finder.condition_count * 2 - 1, 2)
        for _ in range(samples):
            next_dt = self._dt_finder.calc_next(next_dt)
            yield f'{indent:s}     - {next_dt}'


class VirtualMeterOperation(SupportsDateTimeAction):
    def __init__(self, dt_finder: DateTimeFinder, start_now: bool) -> None:
        super().__init__(dt_finder, start_now)
        self.last_value: float | None = None
        self.offset: float | None = None

    @override
    def on_first_value(self, value, info: SmlValueInfo):
        self.last_value = value
        self.offset = value
        return self.process_value(value, info)

    @override
    def process_value(self, value: float | None, info: SmlValueInfo) -> float | None:
        if value is None:
            return None

        if self.after_next_reset():
            self.offset = self.last_value

        self.last_value = value

        if (offset := self.offset) is None:
            return None

        return value - offset

    def __repr__(self) -> str:
        return (f'<VirtualMeter: next_reset={self._next_reset} offset={self.offset} '
                f'last_value={self.last_value} at 0x{id(self):x}>')

    @override
    def describe(self, indent: str = '') -> Generator[str, None, None]:
        yield f'{indent:s}- Virtual Meter:'
        yield f'{indent:s}    Offset: {self.offset}'
        yield from super().describe(indent)


class MaxValueOperation(SupportsDateTimeAction):
    def __init__(self, dt_finder: DateTimeFinder, start_now: bool) -> None:
        super().__init__(dt_finder, start_now)
        self.max_value: float | None = None

    @override
    def on_first_value(self, value, info: SmlValueInfo):
        self.max_value = value
        return self.process_value(value, info)

    @override
    def process_value(self, value: float | None, info: SmlValueInfo) -> float | None:
        if value is None:
            return None

        if self.after_next_reset():
            self.max_value = value
            return value

        if self.max_value is None or value <= self.max_value:
            return None

        self.max_value = value
        return value

    @override
    def describe(self, indent: str = '') -> Generator[str, None, None]:
        yield f'{indent:s}- Max Value:'
        yield f'{indent:s}    max: {self.max_value}'
        yield from super().describe(indent)


class MinValueOperation(SupportsDateTimeAction):
    def __init__(self, dt_finder: DateTimeFinder, start_now: bool) -> None:
        super().__init__(dt_finder, start_now)
        self.min_value: float | None = None

    @override
    def on_first_value(self, value, info: SmlValueInfo):
        self.min_value = value
        return self.process_value(value, info)

    @override
    def process_value(self, value: float | None, info: SmlValueInfo) -> float | None:
        if value is None:
            return None

        if self.after_next_reset():
            self.min_value = value
            return value

        if self.min_value is None or value >= self.min_value:
            return None

        self.min_value = value
        return value

    @override
    def describe(self, indent: str = '') -> Generator[str, None, None]:
        yield f'{indent:s}- Min Value:'
        yield f'{indent:s}    min: {self.min_value}'
        yield from super().describe(indent)



from __future__ import annotations

from pathlib import Path

import setuptools  # type: ignore


# Load version number without importing HABApp
def load_version() -> str:
    version: dict[str, str] = {}
    with open('src/sml2mqtt/__version__.py') as fp:
        exec(fp.read(), version)
    assert version['__version__'], version
    return version['__version__']


def load_req() -> list[str]:
    req_file = Path(__file__).with_name('requirements_setup.txt')
    with req_file.open() as f:
        return f.readlines()


__version__ = load_version()

print(f'Version: {__version__}')
print('')

# When we run tox tests we don't have these files available so we skip them
readme = Path(__file__).with_name('readme.md')
long_description = ''
if readme.is_file():
    with readme.open('r', encoding='utf-8') as fh:
        long_description = fh.read()

setuptools.setup(
    name='sml2mqtt',
    version=__version__,
    author='spaceman_spiff',
    # author_email="",
    description='A sml (Smart Message Language) energy meter to MQTT bridge. '
                'Can read from serial ports or http (e.g. Tibber Pulse).',
    keywords=[
        'mqtt',
        'sml',
        'Smart Message Language',
        'energy meter',
        'tibber'
    ],
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/spacemanspiff2007/sml2mqtt',
    project_urls={
        'GitHub': 'https://github.com/spacemanspiff2007/sml2mqtt',
    },
    packages=setuptools.find_packages(where='src', exclude=['tests*']),
    package_dir={'': 'src'},
    python_requires='>=3.10',
    install_requires=load_req(),
    classifiers=[
        'Development Status :: 4 - Beta',
        'Framework :: AsyncIO',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Home Automation'
    ],
    entry_points={
        'console_scripts': [
            'sml2mqtt = sml2mqtt.__main__:main'
        ]
    }
)



import logging

from easyconfig import BaseModel
from pydantic import Field, field_validator


class LoggingSettings(BaseModel):
    level: str = Field('INFO', description='Log level')
    file: str = Field('sml2mqtt.log', description='Log file path (absolute or relative to config file) or "stdout"')

    @field_validator('level')
    def validate_logging(cls, value):
        if value not in logging._nameToLevel:
            msg = f'Level must be one of {", ".join(logging._nameToLevel)}'
            raise ValueError(msg)
        return value

    def set_log_level(self) -> int:
        level = logging._nameToLevel[self.level]
        logging.getLogger().setLevel(level)
        return level



from __future__ import annotations

from typing import TYPE_CHECKING

from sml2mqtt.errors import RequiredObisValueNotInFrameError, UnprocessedObisValuesReceivedError


if TYPE_CHECKING:
    from collections.abc import Generator

    from sml2mqtt.const import SmlFrameValues
    from sml2mqtt.sml_value.sml_value import SmlValue


class SmlValues:
    def __init__(self) -> None:
        self._processed_ids: frozenset[str] = frozenset()
        self._skipped_ids: frozenset[str] = frozenset()
        self._all_ids: frozenset[str] = frozenset()
        self._values: tuple[SmlValue, ...] = ()

    def __repr__(self) -> str:
        return (
            f'<{self.__class__.__name__:s} '
            f'processed={",".join(self._processed_ids):s}, '
            f'skipped={",".join(self._skipped_ids):s}>'
        )

    def set_skipped(self, *obis_ids: str):
        self._skipped_ids = frozenset(obis_ids)
        self._all_ids = self._processed_ids | self._skipped_ids
        return self

    def add_value(self, value: SmlValue):
        self._processed_ids = self._processed_ids.union((value.obis, ))
        self._all_ids = self._processed_ids | self._skipped_ids
        self._values = (*self._values, value)
        return self

    def process_frame(self, frame: SmlFrameValues):
        for value in self._values:
            value.process_frame(frame)

        obis_in_frame = frame.obis_ids()

        # Not all obis processed
        if obis_left := obis_in_frame - self._all_ids:
            entries_left = [frame.get_value(_obis) for _obis in sorted(obis_left)]
            raise UnprocessedObisValuesReceivedError(*entries_left)

        # Processed obis not in frame
        if obis_missing := self._processed_ids - obis_in_frame:
            raise RequiredObisValueNotInFrameError(*sorted(obis_missing))

    def describe(self) -> Generator[str, None, None]:
        yield f'Skipped: {", ".join(sorted(self._skipped_ids))}'
        yield ''
        for value in self._values:
            yield from value.describe()



from collections.abc import Generator
from time import monotonic
from typing import Final

from typing_extensions import override

from sml2mqtt.const import DurationType, get_duration
from sml2mqtt.sml_value.base import SmlValueInfo, ValueOperationBase
from sml2mqtt.sml_value.operations._helper import format_period


class OnChangeFilterOperation(ValueOperationBase):
    def __init__(self) -> None:
        self.last_value: int | float | str | None = None

    @override
    def process_value(self, value: float | None, info: SmlValueInfo) -> float | None:
        if value is None:
            return None

        if self.last_value == value:
            return None

        self.last_value = value
        return value

    def __repr__(self) -> str:
        return f'<OnChangeFilter at 0x{id(self):x}>'

    @override
    def describe(self, indent: str = '') -> Generator[str, None, None]:
        yield f'{indent:s}- On Change Filter'


class RangeFilterOperation(ValueOperationBase):
    # noinspection PyShadowingBuiltins
    def __init__(self, min_value: float | None, max_value: float | None, limit_values: bool = True) -> None:
        self.min_value: Final = min_value
        self.max_value: Final = max_value
        self.limit_values: Final = limit_values

    @override
    def process_value(self, value: float | None, info: SmlValueInfo) -> float | None:
        if value is None:
            return None

        if (min_value := self.min_value) is not None and value < min_value:
            return min_value if self.limit_values else None

        if (max_value := self.max_value) is not None and value > max_value:
            return max_value if self.limit_values else None

        return value

    def __repr__(self) -> str:
        return (f'<RangeFilter: min={self.min_value} max={self.max_value} '
                f'limit_values={self.limit_values} at 0x{id(self):x}>')

    @override
    def describe(self, indent: str = '') -> Generator[str, None, None]:
        yield f'{indent:s}- Range Filter:'
        if self.min_value is not None:
            yield f'{indent:s}    min: {self.min_value}'
        if self.max_value is not None:
            yield f'{indent:s}    max: {self.max_value}'
        yield f'{indent:s}    limit to min/max: {self.limit_values}'


class DeltaFilterOperation(ValueOperationBase):
    def __init__(self, min_value: int | float | None = None, min_percent: int | float | None = None) -> None:
        self.min_value: Final = min_value
        self.min_percent: Final = min_percent

        self.last_value: int | float = -1_000_000_000

    @override
    def process_value(self, value: float | None, info: SmlValueInfo) -> float | None:
        if value is None:
            return None

        last_value = self.last_value

        diff = abs(value - last_value)

        if (delta_min := self.min_value) is not None and diff < delta_min:
            return None

        if (min_percent := self.min_percent) is not None:  # noqa: SIM102
            # if last value == 0 the percentual change is infinite and we always pass
            if last_value != 0:
                percent = abs(diff / last_value) * 100
                if percent < min_percent:
                    return None

        self.last_value = value
        return value

    def __repr__(self) -> str:
        return f'<DeltaFilter: min={self.min_value} min_percent={self.min_percent} at 0x{id(self):x}>'

    @override
    def describe(self, indent: str = '') -> Generator[str, None, None]:
        yield f'{indent:s}- Delta Filter:'
        if self.min_value:
            yield f'{indent:s}    Min  : {self.min_value}'
        if self.min_percent:
            yield f'{indent:s}    Min %: {self.min_percent}'


class SkipZeroMeterOperation(ValueOperationBase):

    @override
    def process_value(self, value: float | None, info: SmlValueInfo) -> float | None:
        if value is None or value < 0.1:
            return None
        return value

    def __repr__(self) -> str:
        return f'<SkipZeroMeter at 0x{id(self):x}>'

    @override
    def describe(self, indent: str = '') -> Generator[str, None, None]:
        yield f'{indent:s}- Zero Meter Filter'


class ThrottleFilterOperation(ValueOperationBase):
    def __init__(self, period: DurationType) -> None:
        self.period: Final = get_duration(period)
        self.last_time: float = -1_000_000_000

    @override
    def process_value(self, value: float | None, info: SmlValueInfo) -> float | None:
        if value is None:
            return None

        now = monotonic()
        if self.last_time + self.period > now:
            return None

        self.last_time = now
        return value

    def __repr__(self) -> str:
        return f'<ThrottleFilter: {self.period}s at 0x{id(self):x}>'

    @override
    def describe(self, indent: str = '') -> Generator[str, None, None]:
        yield f'{indent:s}- Throttle Filter: {format_period(self.period)}'
